package ca.mun.cs.av;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.Sprite;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.math.Affine2;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.FloatArray;
import com.badlogic.gdx.utils.Disposable;

/**
 * A graphical tree that grows up and branches outwards in a fractal pattern.  The level of growth
 * corresponds to the level of recursion.  Water drops that strike the leaves of the tree accumulate
 * over time and trigger growth.  Once the maximum growth level is reached, berries appear.  Yum!
 *
 * @author Andrew Vardy <av@mun.ca>
 */
class RecursiveTree implements Tree {
    private int baseX, baseY;
    private Texture stickLeftImage, stickRightImage, blueberryImage;
    private Sprite stickLeft, stickRight, blueberry;
    private float stickLeftHeight, stickLeftWidth, stickRightWidth, stickRightHeight;

    private FloatArray leafPoints;

    // Number of water drops to hit the tree since the last call to 'update'.
    private int drops = 0;

    // The tree's level of growth.
    private int growthLevel = 1;

    public static final int MAX_GROWTH_LEVEL = 9;

    // Parameters governing the shape of the tree.
    public static final float FIRST_BRANCH_SCALE = 0.6f;
    public static final float SECOND_BRANCH_SCALE = 0.8f;
    public static final int LEAVES_PER_BERRY = 10;

    // How many raindrops are required to cause the tree to grow?
    public static final int DROPS_FOR_GROWTH = 100;

    RecursiveTree(int baseX, int baseY) {
        this.baseX = baseX;
        this.baseY = baseY;
        stickLeftImage = new Texture(Gdx.files.internal("stick_left.png"));
        stickRightImage = new Texture(Gdx.files.internal("stick_right.png"));
        blueberryImage = new Texture(Gdx.files.internal("blueberry.png"));
        stickLeft = new Sprite(stickLeftImage);
        stickRight = new Sprite(stickRightImage);
        blueberry = new Sprite(blueberryImage);
        stickLeftWidth = stickLeft.getWidth();
        stickLeftHeight = stickLeft.getHeight();
        stickRightWidth = stickRight.getWidth();
        stickRightHeight = stickRight.getHeight();

        update();
    }

    public void draw(SpriteBatch batch) {

        // An affine transform is used to represent translation, rotation, and scaling operations.
        Affine2 transform = new Affine2();

        // Initial translation and rotation, bringing us to the base of the tree, pointed upwards.
        transform.translate(baseX, baseY);
        transform.rotate(90.0f);

        // Keep track of the points on the drawn tree where leaves exist (in this universe, it
        // is the leaves that absorb water into the tree).
        leafPoints = new FloatArray();

        recursiveDraw(batch, transform, growthLevel);

        // Ready for berries yet?
        if (growthLevel == MAX_GROWTH_LEVEL) {
            for (int i=0; i<leafPoints.items.length-1; i += 2 * LEAVES_PER_BERRY) {
                batch.draw(blueberry, leafPoints.items[i], leafPoints.items[i + 1]);
            }
        }
    }

    private void recursiveDraw(SpriteBatch batch, Affine2 transform, int level) {
        if (level == 0)
            // Base case: do nothing.
            return;

        // Store the current transform state for use below.
        Affine2 savedTransform = new Affine2(transform);

        drawBranch(batch, transform);

        // Translate to the first branching point
        transform.translate(stickLeftWidth * 0.86f, 0);

        // Recursively draw the first branch (and its sub-tree)
        transform.rotate(30.0f);
        transform.scale(FIRST_BRANCH_SCALE, FIRST_BRANCH_SCALE);
        addLeafPoint(transform);
        recursiveDraw(batch, transform, level - 1);

        // Reposition to second branching point by restoring the saved transform.
        transform = savedTransform;
        transform.translate(stickLeftWidth * 0.55f, 0);

        // Recursively draw the second branch (and its sub-tree)
        transform.rotate(-30.0f);
        transform.scale(SECOND_BRANCH_SCALE, SECOND_BRANCH_SCALE);
        addLeafPoint(transform);
        recursiveDraw(batch, transform, level - 1);
    }

    private void drawBranch(SpriteBatch batch, Affine2 transform) {
        // Draw the current branch.  We draw it as two halves because transformations such as
        // rotation are made with respect to the lower left corner of the image.
        batch.draw(stickLeft, stickLeftWidth, stickLeftHeight, transform);
        transform.scale(1f, -1f);
        batch.draw(stickRight, stickRightWidth, stickRightHeight, transform);
        transform.scale(1f, -1f);
    }

    private void addLeafPoint(Affine2 transform) {
        Vector2 A = new Vector2(0, 0);
        transform.applyTo(A);
        leafPoints.add(A.x, A.y);
    }

    public void addDrop() {
        drops++;
    }

    /**
     * Updates the tree's state by checking whether the growth level should be advanced based on
     * the amount of water absobed.
     */
    public void update() {
        if (growthLevel < MAX_GROWTH_LEVEL && drops >= DROPS_FOR_GROWTH) {
            growthLevel += 1;
            drops = 0;
        }
        //Gdx.app.log("RecursiveTree", "growthLevel: " + growthLevel);
    }

    /**
     * @pre leafPoints already contains the set of all leaf points.
     * @return Whether (x, y) is within a threshold distance of any leaf points.
     */
    public boolean closeToAnyLeafPoints(float x, float y) {
        for (int i=0; i<leafPoints.items.length-1; i += 2) {
            float dx = x - leafPoints.items[i];
            float dy = y - leafPoints.items[i + 1];
            double dist = Math.sqrt(dx*dx + dy*dy);

            if (dist < 10) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void dispose() {
        stickLeftImage.dispose();
        stickRightImage.dispose();
    }
}
