package ca.mun.cs.av;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.Sprite;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.math.Affine2;
import com.badlogic.gdx.utils.Disposable;

/**
 * A simple tree with two branches.  A stepping stone towards the more sophisticated RecursiveTree class.
 *
 * @author Andrew Vardy <av@mun.ca>
 */
class SimpleTree implements Tree {
    private int baseX, baseY;
    private Texture stickLeftImage, stickRightImage;
    private Sprite stickLeft, stickRight;
    private float stickLeftHeight, stickLeftWidth, stickRightWidth, stickRightHeight;

    // Parameters governing the shape of the tree.
    public static final float FIRST_BRANCH_SCALE = 0.6f;
    public static final float SECOND_BRANCH_SCALE = 0.8f;

    SimpleTree(int baseX, int baseY) {
        this.baseX = baseX;
        this.baseY = baseY;
        stickLeftImage = new Texture(Gdx.files.internal("stick_left.png"));
        stickRightImage = new Texture(Gdx.files.internal("stick_right.png"));
        stickLeft = new Sprite(stickLeftImage);
        stickRight = new Sprite(stickRightImage);
        stickLeftWidth = stickLeft.getWidth();
        stickLeftHeight = stickLeft.getHeight();
        stickRightWidth = stickRight.getWidth();
        stickRightHeight = stickRight.getHeight();
    }

    public void draw(SpriteBatch batch) {
        // An affine transform is used to represent translation, rotation, and scaling operations.
        Affine2 transform = new Affine2();

        // Initial translation and rotation, bringing us to the base of the tree, pointed upwards.
        transform.translate(baseX, baseY);
        transform.rotate(90.0f);

        // Store the current transform state for use below.
        Affine2 savedTransform = new Affine2(transform);

        drawBranch(batch, transform);

        // Translate to the first branching point
        transform.translate(stickLeftWidth * 0.86f, 0);

        // Draw the first branch
        transform.rotate(30.0f);
        transform.scale(FIRST_BRANCH_SCALE, FIRST_BRANCH_SCALE);
        drawBranch(batch, transform);

        // Reposition to second branching point by restoring the saved transform.
        transform = savedTransform;
        transform.translate(stickLeftWidth * 0.55f, 0);

        // Draw the second branch
        transform.rotate(-30.0f);
        transform.scale(SECOND_BRANCH_SCALE, SECOND_BRANCH_SCALE);
        drawBranch(batch, transform);
    }

    private void drawBranch(SpriteBatch batch, Affine2 transform) {
        // Draw the current branch.  We draw it as two halves because transformations such as
        // rotation are made with respect to the lower left corner of the image.
        batch.draw(stickLeft, stickLeftWidth, stickLeftHeight, transform);
        transform.scale(1f, -1f);
        batch.draw(stickRight, stickRightWidth, stickRightHeight, transform);
        transform.scale(1f, -1f);
    }

    /**
     * The following methods are just placeholders.  Real implementations in RecursiveTree.
     */
    public void addDrop() {}
    public void update() {}
    public boolean closeToAnyLeafPoints(float x, float y) { return false; }

    @Override
    public void dispose() {
        stickLeftImage.dispose();
        stickRightImage.dispose();
    }
}
