package ca.mun.cs.av;

import java.util.Iterator;

import com.badlogic.gdx.ApplicationAdapter;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input.Keys;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.OrthographicCamera;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.graphics.g2d.Sprite;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.TimeUtils;

/**
 * A relaxing "game" involving two growing trees.  The user's job is to move the cloud to rain
 * on these trees, thus encouraging them to grow.
 *
 * @author Andrew Vardy <av@mun.ca>
 */
public class ZenGardenGame extends ApplicationAdapter {
	private Texture cloudImage, dropImage;
	private SpriteBatch batch;
	private OrthographicCamera camera;
	private Sprite cloud;
	private Array<Sprite> raindrops;
	private long lastDropTime;
	private Array<Tree> trees;

	// The width and height of the screen --- assumed not to change (otherwise define resize).
	private int width, height;

	@Override
	public void create() {
		// load the images for the droplet and the stick
        cloudImage = new Texture(Gdx.files.internal("cloud.png"));
		dropImage = new Texture(Gdx.files.internal("droplet.png"));

		width = Gdx.graphics.getWidth();
		height = Gdx.graphics.getHeight();

		// create the camera and the SpriteBatch
		camera = new OrthographicCamera();
		camera.setToOrtho(false, width, height);
		batch = new SpriteBatch();

		// create a Sprite to logically represent the cloud
		cloud = new Sprite(cloudImage);
		cloud.setX(width / 2 - cloud.getWidth() / 2); // center the cloud horizontally
		cloud.setY(height - cloud.getHeight()); // bottom left corner of the cloud is 20 pixels above the bottom screen edge

		// create the raindrops array and spawn the first raindrop
		raindrops = new Array<Sprite>();
		spawnRaindrop();

		trees = new Array<Tree>();

        // Uncomment the following for the simplified 2-level, non-growing tree.
		//trees.add(new SimpleTree(width / 4, 0));
		//trees.add(new SimpleTree(3 * width / 4, 0));

        // Use "RecursiveTree" to get the complete, functional tree growth.
        trees.add(new RecursiveTree(width / 4, 0));
        trees.add(new RecursiveTree(3 * width / 4, 0));
	}

	private void spawnRaindrop() {
		Sprite raindrop = new Sprite(dropImage);
		float cloudX = cloud.getX();
		raindrop.setX(MathUtils.random(cloudX, cloudX + cloud.getWidth()));
		raindrop.setY(height - cloud.getHeight());
		raindrops.add(raindrop);
		lastDropTime = TimeUtils.nanoTime();
	}

	@Override
	public void render() {
		Gdx.gl.glClearColor(0.91f, 0.95f, 1.0f, 1);
		Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);

		camera.update();

		batch.setProjectionMatrix(camera.combined);

		// begin a new batch and draw the cloud and
		// all drops
		batch.begin();
		cloud.draw(batch);
		for(Sprite raindrop: raindrops)
			raindrop.draw(batch);
		for (Tree tree : trees)
			tree.draw(batch);
		batch.end();

		// process user input
		if(Gdx.input.isTouched()) {
			Vector3 touchPos = new Vector3();
			touchPos.set(Gdx.input.getX(), Gdx.input.getY(), 0);
			camera.unproject(touchPos);
			cloud.setX(touchPos.x - cloud.getWidth() / 2);
		}
		if(Gdx.input.isKeyPressed(Keys.LEFT)) cloud.translateX( -400 * Gdx.graphics.getDeltaTime());
		if(Gdx.input.isKeyPressed(Keys.RIGHT)) cloud.translateX( 400 * Gdx.graphics.getDeltaTime());

		// make sure the cloud stays within the screen bounds
		if(cloud.getX() < 0) cloud.setX(0);
		if(cloud.getX() > width - cloud.getWidth()) cloud.setX(width - cloud.getWidth());

		// check if we need to create a new raindrop
		if(TimeUtils.nanoTime() - lastDropTime > 10000000) spawnRaindrop();

		// move the raindrops, remove any that are beneath the bottom edge of
		// the screen.
		Iterator<Sprite> iter = raindrops.iterator();
		while(iter.hasNext()) {
			Sprite raindrop = iter.next();
			raindrop.translateY(-200 * Gdx.graphics.getDeltaTime());
			if(raindrop.getY() + raindrop.getHeight() < 0) iter.remove();

			for (Tree tree : trees)
				if(tree.closeToAnyLeafPoints(raindrop.getX(), raindrop.getY())) {
					tree.addDrop();
					iter.remove();
				}
		}

		for (Tree tree : trees)
			tree.update();
	}

	@Override
	public void dispose() {
		// dispose of all the native resources
		dropImage.dispose();
		cloudImage.dispose();
		for (Tree tree : trees)
			tree.dispose();
		batch.dispose();
	}
}